

type atom = string
type sym = string
type var = Var.var
type atom_sort = string
type data_sort = string

type sort = AtomSt of atom_sort
  | DataSt of data_sort
  | UnitSt
  | PairSt of sort * sort
  | AbsSt of string * sort
  | ListSt of sort

type arity = sort * data_sort

type sym_kind = AtomSK 
  | DataSK 
  | AtomK of atom_sort 
  | SymK of arity 
  | PropK of sort

type sg = (string * sym_kind) list

type perm
val id : perm
val transpose : atom -> atom -> perm
val mk_perm : (atom*atom) list -> perm

type term = Atom of atom
  | Unit   
  | Pair of term * term
  | App of sym * term
  | Abs of atom * term
  | Susp of perm * var


type equations = (term*term) list
type subst
val mk_subst : (var * term) list -> subst
val empty : subst
val join : subst -> subst -> subst
val lookup : subst -> var -> term option
type constr = (atom * term) list

type prop = TermProp of sym * term
  | FreshProp of atom * term

type clause = prop * constr * prop list

type decl = AtomDecl of atom_sort
  | DataDecl of data_sort
  | PropDecl of sym * sort
  | SymDecl of sym * arity
  | ClauseDecl of clause 
  | Query of prop * prop list
  | UseDirective of string
  | TraceDirective of int

val apply_subst : subst -> term -> term
val apply_subst_prop : subst -> prop -> prop
val apply_perm : perm -> term -> term
val apply_perm_atom : perm -> atom -> atom
val apply_perm_prop : perm -> prop -> prop
val invert_perm : perm -> perm
val domain_perm : perm -> atom list
val disagreement_set : perm -> perm -> atom list

val pp_term : out_channel -> term -> unit;;
val pp_sort : out_channel -> sort -> unit;;
val pp_perm : out_channel -> perm -> unit;;
val pp_subst : out_channel -> subst -> unit;;
val pp_clause : out_channel -> clause -> unit;;
val pp_constr : out_channel -> constr -> unit;;
val pp_decl : out_channel -> decl -> unit;;
val pp_prop : out_channel -> prop -> unit;;

val sort_to_string : sort -> string;;

val fvs : term -> var list;;
val fvs_prop : prop -> var list;;

val rename : prop -> prop list -> prop * prop list;;
val atom_occurs_in : atom -> term -> bool
