{
open Parser
let lineno = ref 1
}

let alpha = ['A'-'Z' 'a'-'z']
let upper = ['A'-'Z']
let lower = ['a'-'z']
let digit = ['0'-'9']
let varname = (upper|'_')(alpha|digit|'_'|''')*
let aname = '@'(alpha|digit|'_'|''')*
let name = (lower)(alpha|digit|'_'|''')*
let filename = '"'(alpha|digit|'_'|'.'|'/'|'\\'|':')+'"'
let ws = [' ' '\t']

rule lex =
  parse
        '/''/'[^'\n']*'\n' { lineno := !lineno + 1; lex lexbuf }
      | ['\n']     { lineno := !lineno + 1; lex lexbuf }
      | ws+        { lex lexbuf }
      | digit+     { INT(int_of_string(Lexing.lexeme lexbuf)) }
      | "|"        { BAR }
      | "["        { LBRACK }
      | "]"        { RBRACK }
      | "."        { DOT }
      | ","        { COMMA }
      | "->"       { ARROW }
      | "<-"       { LARROW }
      | "?-"       { QUERY } 
      | "<"        { LANGLE }
      | ">"        { RANGLE }
      | ":"        { COLON }
      | ";"        { SEMI }
      | "("        { LPAREN }
      | ")"        { RPAREN }
      | "*"        { TIMESTY }
      | "#"        { HASH }
      | "/"        { SLASH }
      | "unit"     { UNITTY }
      | "atom"     { ATOMSORT }
      | "data"     { DATASORT }
      | "list"     { LISTTY }
      | "prop"     { PROPSORT }
      | "#use"     { USE }
      | "#trace"   { TRACE }
      | filename   { let s = Lexing.lexeme lexbuf in
	             FILENAME( String.sub s 1 (String.length(s) - 2) )}
      | name       { NAME(Lexing.lexeme lexbuf) }
      | aname      { let s = Lexing.lexeme lexbuf in
	             ANAME(String.sub s 1 (String.length (s) - 1))}
      | varname    { VAR(Lexing.lexeme lexbuf) }
      | eof        { EOF }

{
}
