%{
open Absyn
exception Eof
%}

%token <int> INT
%token <string> VAR NAME ANAME FILENAME
%token UNITTY TIMESTY LANGLE RANGLE LPAREN RPAREN DOT 
%token LARROW ARROW COLON ATOMSORT DATASORT PROPSORT QUERY COMMA
%token BAR LBRACK RBRACK LISTTY HASH SLASH
%token EOF USE  SEMI TRACE

%type <Absyn.decl list> parse
%type <Absyn.term> tm parse_tm
%type <Absyn.term list> comma_tms
%type <Absyn.sort> tp parse_tp
%type <Absyn.decl list> parse_line

%right TIMESTY
%nonassoc LISTTY
%nonassoc RANGLE

%start parse
%start parse_tm
%start parse_tp
%start parse_line

%%

parse: decls EOF           { List.rev $1 }

parse_line : decl {[$1]}
| EOF { [] }

decls : decls decl { $2 :: $1 }
      |            { [] }
      ;

decl  : NAME COLON ATOMSORT DOT { AtomDecl($1) }
      | NAME COLON DATASORT DOT { DataDecl($1) }
      | NAME COLON tp ARROW NAME DOT { SymDecl($1,($3,$5)) }
      | NAME COLON NAME DOT { SymDecl($1,(UnitSt,$3)) }
      | NAME COLON tp ARROW PROPSORT DOT { PropDecl($1,$3) }
      | NAME COLON PROPSORT DOT { PropDecl($1,UnitSt) }
      | prop SLASH constrs LARROW comma_props DOT { ClauseDecl($1,$3, List.rev $5) }
      | prop LARROW comma_props DOT { ClauseDecl($1,[], List.rev $3) }
      | prop DOT { ClauseDecl($1,[],[]) }
      | prop QUERY comma_props DOT  { Query($1,List.rev $3) }
      | USE FILENAME DOT { UseDirective($2) }
      | TRACE INT DOT { TraceDirective ($2) }
	 
parse_tm : tm {$1}

prop : NAME { TermProp($1,Unit) }
      | NAME tm {TermProp ($1,$2) }
      | constr { let (x,y) = $1 in FreshProp (x,y) }
;


constr : NAME HASH tm {($1, $3)}
constrs : constr { [$1] }
      | constrs COMMA constr {$3 :: $1}

tm : ANAME {Atom ($1)}
      | VAR {Susp(id,Var.mkvar' $1)}
      | LANGLE NAME RANGLE tm {Abs($2,$4)}
      | LPAREN comma_tms COMMA tm RPAREN {List.fold_right (fun x y -> Pair(x,y)) (List.rev $2) $4}
      | LBRACK comma_tms RBRACK {List.fold_right (fun x y -> Pair(x,y)) (List.rev $2) Unit}
      | LBRACK comma_tms BAR tm RBRACK {List.fold_right (fun x y -> Pair(x,y)) (List.rev $2) $4}
      | NAME tm {App($1,$2)}
      | NAME {App ($1,Unit)}
      | LPAREN tm RPAREN {$2}
;

comma_tms :   { [] }
      | tm { [$1] }
      | comma_tms COMMA tm { $3 :: $1}
;
comma_props :   { [] }
      | prop { [$1] }
      | comma_props COMMA prop { $3 :: $1}
;

parse_tp : tp {$1}

tp :      tp TIMESTY tp {PairSt($1,$3)}
        | UNITTY { UnitSt }
	| tp LISTTY {ListSt $1 }
	| ANAME { AtomSt($1) }
	| NAME { DataSt($1) }
	| LANGLE NAME RANGLE tp { AbsSt($2,$4) }
	| LPAREN tp RPAREN { $2 }
        ;
