open Absyn;;
open Unify;;

let trace = ref 0;;

let do_trace i f = if !trace >= i then f();;

let atom = ref 0;;
let new_atom a = 
  let i = !atom in
  let _ = atom := !atom + 1 in
  a^(string_of_int i);;
let reset_atom () = atom := 0;;

exception Fail;;
(* TODO: Generate freshness constraints for cs that should be present in 
   solution.  *)

type stack = (prop list * clause list * subst * constr) list

let solve decls stk = 
  let rec s' stk = 
    match stk with 
      [] -> 
	raise Fail
    | ([], _, subst, constr)::stk -> 
	(subst,constr,stk)
    | ((FreshProp(a,tm) as p)::gs, [], subst, constr)::stk -> 
	(try let newconstr = unify_fresh ((a,tm)::constr) in
	s'((gs, decls, subst, newconstr)::stk)
	with Unify msg -> 
	  do_trace 3 (fun () -> 
	    print_string "goal ";
	    pp_prop stdout p;
	    print_string " fails because ";
	    print_string msg;
	    print_string "\n");
	  s' stk)
    | ((TermProp(_) as p)::gs, [], subst, constr)::stk -> 
	do_trace 1 (fun () -> print_string "Backtracking from ";
	                      pp_prop stdout p; 
	                      print_string "\n");
	s' stk
    | (g::gs, ((p,cs,ps) as clause)::clauses, subst, constr)::stk ->
	(try let p,ps = Absyn.rename p ps in
	let freshatoms = List.map (fun (a,_) -> a) cs in
	let perm = mk_perm(List.map (fun a -> (a,new_atom(a))) freshatoms) in
	let p,ps = apply_perm_prop perm p, List.map (apply_perm_prop perm) ps in
        let newsubst, newconstr = unify_prop g p constr in
	do_trace 1 
	  (fun () -> print_string "Matching clause ";
	    pp_prop stdout g;
	    print_string " with\n\t";
	    pp_clause stdout clause;
	    print_string "\n");
	do_trace 2 
	  (fun () -> 
	    pp_subst stdout newsubst;
	    pp_constr stdout newconstr);
	s'(((List.map (apply_subst_prop newsubst) (ps@gs)),
	   decls,
	   join newsubst subst, newconstr)::
	   (g::gs, clauses, subst, constr)::stk)
	with Unify msg -> 
	  do_trace 3 (fun () -> 
	    print_string "goal ";
	    pp_prop stdout g;
	    print_string " and clause ";
	    pp_clause stdout clause; 
	    print_string " fail to resolve because ";
	    print_string msg;
	    print_string "\n");
	  s' ((g::gs, clauses, subst,constr)::stk))
  in s' stk
