open Absyn;;



exception Unify of string

let rec unify_fresh fs = 
  match fs with
  | (_,Unit)::fs -> unify_fresh fs
  | (a,Atom b)::fs -> 
      if a = b then raise (Unify ("freshness check for "^a^" failed"))
      else unify_fresh fs
  | (a,Abs(b,e))::fs -> 
      if a = b then unify_fresh fs
      else unify_fresh ((a,e)::fs)
  | (a,App(f,e))::fs -> 
      unify_fresh ((a,e)::fs)
  | (a, Pair(e1,e2)) :: fs -> 
      unify_fresh ((a,e1)::(a,e2)::fs)
  | (a,Susp(p,x))::fs -> 
      (apply_perm_atom (invert_perm p) a, Susp(Absyn.id,x))::(unify_fresh fs)
  | [] -> []

let do_susp p1 p2 x fs = 
  let ds = disagreement_set p1 p2 in
  (List.map (fun a -> (a,Susp(id,x))) ds)@fs

let rec occurs_check x t = 
  match t with
    Unit -> ()
  | Atom a -> ()
  | Pair(e1,e2) -> occurs_check x e1; occurs_check x e2
  | Abs(a,e) -> occurs_check x e
  | App(f,e) -> occurs_check x e
  | Susp(_,y) -> if Var.eq x y then raise (Unify ("occurs check failed for "^(Var.to_string x^" and "^Var.to_string y)))
	else ()

let rec do_subst t p x es fs = 
  occurs_check x t;
  let s = mk_subst[(x,apply_perm (invert_perm p) t)] in
  let es' = List.map (fun (e1,e2) -> apply_subst s e1, apply_subst s e2) es in
  let fs' = List.map (fun (a,e) -> a,apply_subst s e) fs in
  let s',f' = unify es' fs' in
  join s s',f'

and unify es fs = 
  match es with
    (Atom a, Atom b) :: es -> 
      if a = b then unify es fs
      else raise ( Unify ("atoms "^a^" and "^b^" differ"))
  | (Unit,Unit)::es -> unify es fs
  | (Abs(a,e), Abs(b,e'))::es -> 
      if a = b then unify ((e,e')::es) fs
      else 
	unify ((e,apply_perm (transpose a b) e')::es) ((a,e')::fs)
  | (App(f,e1),App(g,e2))::es -> 
      if f = g then unify ((e1,e2)::es) fs 
      else raise (Unify ("symbols "^f^" and "^g^" differ"))
  | (Pair(e1,e2), Pair(e1',e2'))::es -> 
      unify((e1,e1')::(e2,e2')::es) fs
  | (Susp(p1,x),Susp(p2,y))::es -> 
      if x = y then unify es (do_susp p1 p2 x fs)
      else do_subst (Susp(p1,x)) p2 y es fs
  | (Susp(p,x), t)::es -> do_subst t p x es fs
  | (t,Susp(p,x))::es -> do_subst t p x es fs
	(* freshness cases *)
  | [] -> empty,unify_fresh fs
  | _ -> raise (Unify ("terms are not unifiable"))
	  


let unify_prop p1 p2 constrs = 
  match p1,p2 with
  | TermProp (f,tm), TermProp (f',tm') -> 
      if f = f' 
      then unify [(tm,tm')] constrs
      else raise (Unify "symbol heads don't match")
  | FreshProp (a,tm), FreshProp(a',tm') -> 
      if a = a' 
      then unify [(tm,tm')] constrs 
      else raise (Unify "fresh variables don't match")
  | _ -> raise (Unify "proposition constructors don't match")

